#include <cstdio>
#include <cstring> // memset

#include "cpu.hh"
#include "ppu.hh"
#include "papu.hh"
#include "pads.hh"
#include "clock.h"

#include <unistd.h>
#include <sys/time.h>
int mainTick()
{
    fd_set fds;
    struct timeval tv = {0,0};
    FD_ZERO(&fds);
    FD_SET(0, &fds);
    
    return select(1, &fds, NULL, NULL, &tv) > 0;
}

#include "mappers.hh" // FIRST INCLUSION

class bisqnes
{
private:
    friend mapper;
    
    m6502cpu cpu;
    
    bisqppu ppu;
    bisqpapu papu;
    bisqpads pads;
    
    BYTE ram[0x800];   // Mirrored four times
    BYTE sram[0x2000]; // Battery RAM
    
    BYTE *rom;
    BYTE rom16count;
    BYTE ctrlbyte;
    BYTE vrom8count;
    BYTE mappernum;
    
    BYTE *rompage[8]; // 4k pages (0x1000)
    
    mapper *mapp;
    
public:
    bisqnes() : rom(NULL), mapp(NULL)
    {
    }
    ~bisqnes()
    {
        if(rom)delete[] rom;
    }
    
    void load(const char *fn)
    {
        FILE *fp = fopen(fn, "rb");
        if(!fp)return;

        if(fgetc(fp)=='N'
        && fgetc(fp)=='E'
        && fgetc(fp)=='S'
        && fgetc(fp)=='\32')
        {
            rom16count = fgetc(fp);
            vrom8count = fgetc(fp);
            ctrlbyte = fgetc(fp);
            mappernum = fgetc(fp) | (ctrlbyte>>4);
            fseek(fp, 0x10, SEEK_SET);
            
            if(mappernum >= 0x40)
                mappernum &= 15;
            
            fprintf(stderr, "mappernum %u, 4krom=%u, 4kvrom=%u\n", mappernum, rom16count*4, vrom8count*2);
            map<unsigned, mapper *>::iterator i;
            i = mappers.find(mappernum);
            mapp = (i==mappers.end()) ? NULL : i->second;
            
            if(mapp)
            {
                mapp->vrom4kcount = vrom8count * 2;
                mapp->rom4kcount = rom16count * 4;
            }
            else
            	fprintf(stderr, "(mapper unknown)\n");
            
            if(rom)delete[] rom;
            rom = new BYTE[rom16count * 16384];
            fread(&rom[0], 16384, rom16count, fp);
            
            ppu.resizevrom(vrom8count * 8192);
            fread(ppu.vromptr(), 8192, vrom8count, fp);
        }
        
        fclose(fp);
        
        // ROM pages ok, cpu reset etc.
        poweron();
    }
    
    void poweron()
    {
        unsigned rom4count = rom16count * 4;
        for(unsigned a=0; a<4; ++a)
        {
            rompage[a]   = &rom[             a  * 0x1000];
            rompage[a+4] = &rom[(rom4count-4+a) * 0x1000];
        }
        /*
        for(unsigned a=0; a<8; ++a)
        	fprintf(stderr, "poweron layout: [%u] = %u\n",
        		a, (rompage[a]-&rom[0]) / 0x1000);
        */
        memset(ram, 0, sizeof ram);
        memset(sram, 0, sizeof sram);
        
        if(mapp)
        {
        	mapp->reset();
        	mapp->initialization();
        }
        cpu.Reset();
    }
    void sta_load(const char *fn)
    {
        FILE *fp = fopen(fn, "rb");
        if(!fp)return;
        fread(&sram, 1, 0x2000, fp);
        fread(&ram, 1, 0x800, fp);
        cpu.PC.B.h = fgetc(fp);
        cpu.PC.B.l = fgetc(fp);
        cpu.A = fgetc(fp);
        cpu.P = fgetc(fp);
        cpu.X = fgetc(fp);
        cpu.Y = fgetc(fp);
        cpu.S = fgetc(fp);
        fread(&ppu.spr_ram,1, 0x100, fp);
        fread(&ppu.vram,   1, 0x2000, fp);
        fread(&ppu.namtab, 4, 0x400, fp);
        
        fread(&ppu.palette,2, 0x10, fp);
        
        long pos = ftell(fp);
        
//      for(unsigned a=0; a<8; ++a)
//          rompage[a] = &rom[a * 0x1000];
        
// ST7:  00 00 00 00  02 03 0E 0F   05 08 09 0F  16 17 22 23 
// ST8:  00 00 00 00  02 03 0E 0F   01 08 09 0F  16 17 1A 1B

        fseek(fp, 0, SEEK_END);
        fprintf(stderr, "%ld unknown bytes at sta file:\n", ftell(fp)-pos);
        fseek(fp, pos, SEEK_SET);
        for(int b=0;;++b)
        {
        	int a = fgetc(fp);
        	if(a == EOF)break;
        	if(b==8)fprintf(stderr," ");else if(b==16){b=0;fprintf(stderr,"\n");}
        	fprintf(stderr, "%02X ", a);
        }
        fclose(fp);
        fprintf(stderr, "\nLoaded %s.\n", fn);
    }
public:
    inline void Wr(register fWORD Addr, register fBYTE Value)
    {
        if(Addr < 0x2000) { ram[Addr & 0x7FF] = Value; return; }
        if(Addr < 0x6000)
        {
            if(Addr == 0x4014)
            {
                /* DMA transfer source address */
                Addr = Value * 0x100;
                /* This _should_ move one byte per 6 tick()s,
                 * but it's hard to write.
                 */
                for(unsigned a=0; a<256; ++a)
                    ppu.spr_ram[a] = Rd(Addr++);
                return;
            }
            if(ppu.write(Addr, Value))return;
            if(papu.write(Addr, Value))return;
            if(pads.write(Addr, Value))return;
        }
        else if(Addr < 0x8000)
        {
            if(ctrlbyte&8)
                sram[Addr & 0x1FFF] = Value;
        }
        else
        {
            if(mapp)
                mapp->write(Addr, Value);
            else
                fprintf(stderr, "Unknown mapper %u wrote %02X to %04ZX\n", mappernum, Value, Addr);
        }
    }
    inline BYTE Rd(register fWORD Addr)
    {
        if(Addr >= 0x8000)return rompage[(Addr-0x8000)/0x1000][Addr & 0xFFF];
        if(Addr < 0x2000)return ram[Addr & 0x7FF];
        if(Addr < 0x6000)
        {
            BYTE value = 0;
            if(ppu.read(Addr, value))return value;
            if(papu.read(Addr, value))return value;
            if(pads.read(Addr, value))return value;
        }
        return sram[Addr & 0x1FFF];
    }
    inline BYTE Op(register fWORD Addr) const
    {
        if(Addr >= 0x8000)return rompage[(Addr-0x8000)/0x1000][Addr & 0xFFF];
        if(Addr < 0x2000)return ram[Addr & 0x7FF];
        if(Addr < 0x6000)return 0;
        return sram[Addr & 0x1FFF];
    }
    void run();
};

void bisqnes::run()
{
    fBYTE divider = Clock->CPUclockDivider;
    BYTE keycounter = 0;
    for(;;)
    {
    	if(ppu.tick(divider))
    		cpu.Int(INT_NMI);
    	
        if(cpu.ready())
        {
            cpu.Tick();
            if((!++keycounter) && mainTick())break;
        }
        if(papu.tick())cpu.Int(INT_IRQ);
    }
}

static bisqnes nes;

#include "mappers.cc"

void Wr6502(register fWORD Addr, register fBYTE Value)
{
    nes.Wr(Addr, Value);
}
BYTE Rd6502(register fWORD Addr)
{
    return nes.Rd(Addr);
}
BYTE Op6502(register fWORD Addr)
{
    return nes.Op(Addr);
}

int main(int argc, const char *const *argv)
{
    nes.load(argc > 1 ? argv[1] : "mariobrs.nes");
    if(argc > 2)nes.sta_load(argv[2]);
    nes.run();
    return 0;
}
