#ifndef BisqPAPUinternals_hh
#define BisqPAPUinternals_hh

#include "clock.h"

#define PAPUTICK_60HZ(val) (!((val)&3))    // 00 but no 01 02 03   (1/4)
#define PAPUTICK_120HZ(val) (!((val)&1))   // 00 02 but no 01 03   (1/2)
#define PAPUTICK_240HZ(val) true           // 00 01 02 03          (1/1)

class waveunit
{
public:
    PAIR wavelen;
    WORD wavepos;
public:
    waveunit() : wavepos(0)
    {
        wavelen.W = 0;
    }
    inline void waveunit_write(fBYTE ind, fBYTE value)
    {
        if(ind == 2)wavelen.B.l = value;
        else if(ind == 3)wavelen.B.h = value&7;
    }
    inline WORD getwavelen() const { return wavelen.W; }
    inline bool setwavelen(fWORD len)
    {
        register bool carry;
        if(len > 0x7FF)
        {
            len &= 0x7FF;
            carry = true;
        }
        else
            carry = false;
        
        wavelen.W = len;
        return carry;
    }
};
    // This requires CPU tick
#define BEGIN_WAVETICKER inline void waveunit_tick() { if(!wavepos) {
#define END_WAVETICKER wavepos = wavelen.W; } else --wavepos; }

class lengthunit
{
    fBYTE lengthcounter;
    fBYTE bmask;
    bool lengthdisabled;
    bool chanenabled;
public:
    lengthunit(fBYTE mask = 5) : lengthcounter(0), bmask(mask), lengthdisabled(true), chanenabled(true)
    {
    }
    inline void lengthunit_write(fBYTE ind, fBYTE value)
    {
        if(ind == 3)
        {
            static const BYTE lengthtrans[32]=
            {5,127,10,1,20, 2,40, 3,80, 4,30, 5,7, 6,14, 7,
             6,  8,12,9,24,10,48,11,96,12,36,13,8,14,16,15};
            // Ks papu.txt jos tuon taulukon sislt kiinnostaa
            
            if(chanenabled)
            {
                lengthcounter = lengthtrans[(value >> 3) & 31];
                fprintf(stderr, "(\33[1;33mlcounter=%u\33[m)", lengthcounter);
            }
            else
            	fprintf(stderr, "(\33[1;31mlcounter NOT MODIFIED\33[m)");
        }
        if(ind == 0)
        {
            lengthdisabled = (value & bmask);
            if(lengthdisabled)
            	fprintf(stderr, "(\33[33mlength DISabled at %u\33[m)", lengthcounter);
            else
            	fprintf(stderr, "(\33[33mlength ENabled at %u\33[m)", lengthcounter);
        }
    }
    inline void enable()
    {
        chanenabled = true;
    }
    inline void disable()
    {
        chanenabled = false;
        lengthcounter = 0;
    }
    inline bool lengthunit_silent() const
    {
        return !lengthcounter;
    }
    inline void lengthunit_hztick(BYTE ticktype)
    {
    	if(lengthdisabled)return;
        if(PAPUTICK_60HZ(ticktype))
        {
            if(lengthcounter > 0)--lengthcounter;
        }
    }
};

class decayunit : public lengthunit
{
    fBYTE decaycounter;
    fBYTE decayrate;
    fBYTE decaypos;
    fBYTE decaymode;
public:
    decayunit() : decaycounter(0), decayrate(0), decaypos(0), decaymode(0x10)
    {
    }
    inline void decayunit_write(fBYTE ind, fBYTE value)
    {
        lengthunit_write(ind, value);        
        if(ind == 0)
        {
            decayrate = value & 15;
            decaymode = (value>>4) & 3;
        	fprintf(stderr, "[ drate=%u, dmode=%u ]", decayrate, decaymode);
        }
        else if(ind == 3)
        {
        	fprintf(stderr, "[ dcounter=15 ]");
        	decaycounter = 15;
        }
    }
    inline unsigned getvol() const
    {
        return (decaymode & 1) ? decayrate : decaycounter;
    }
    inline void decayunit_hztick(BYTE ticktype)
    {
        lengthunit_hztick(ticktype);
        
        if(PAPUTICK_240HZ(ticktype))
        {
            if(!decaypos)
            {
                // This counter is for timing; it always runs.
                decaypos = decayrate;                
                
                if(decaycounter > 0)
                    --decaycounter;
                else if(decaymode & 2)
                    decaycounter = 15;
            }
            else
                --decaypos;
        }
    }
};

class squaregenunit : public decayunit, public waveunit
{
    fBYTE sweepdata;
    fBYTE sweeppos;
    bool jammed;
    fBYTE decr;
public:
    fBYTE dutytype;
    fBYTE dutycounter;
public:
    squaregenunit() : sweepdata(0x00), sweeppos(0), jammed(false), decr(0), dutytype(0)
    { 
    }
    inline void set_decr_mode(fBYTE d) { decr = d; }
    inline void write(fBYTE ind, fBYTE value)
    {
        fprintf(stderr, "squaree.write(%u, 0x%02X) - ", ind, value);
        waveunit_write(ind, value);
        decayunit_write(ind, value);
        
        if(ind == 1)
        {
            sweepdata = value;
            fprintf(stderr,
              "{sweepshift %d %s, rate %d, %sabled}",
              value&7, (value&8)?"down":"up", (value>>4)&7, (value&128)?"en":"dis");
        }
        else if(ind == 0)
        {
            if((value >> 6) != dutytype)
                fprintf(stderr, "{duty=%d)\n", value >> 6);
            dutytype = value >> 6;
        }
        jammed = getwavelen() < 8;
        if(jammed)
        	fprintf(stderr, " -> jammed\n");
        else if(silent())
        	fprintf(stderr, " -> silent\n");
        else
        	fprintf(stderr, " -> ok\n");
    }
    void hztick(BYTE ticktype)
    {
        decayunit_hztick(ticktype);
        if(PAPUTICK_120HZ(ticktype))
        {
            if(!sweeppos)
            {
                // Update sweep speed counter
                sweeppos = ((sweepdata >> 4) & 7) + 1;
                if((sweepdata & 0x87) > 0x80 && !silent()) // May not update if silent!
                {
                    unsigned len = getwavelen();
                    if(sweepdata & 15)
                        len -= (len >> (sweepdata & 7)) + decr;
                    else
                        len += (len >> (sweepdata & 7));
                    if(setwavelen(len) || len < 8)
                        jammed = true;
                }
            }
            else
                --sweeppos;
        }
    }
    inline bool silent() const
    {
        return jammed || lengthunit_silent();
    }
    BEGIN_WAVETICKER
        /*if(!silent())*/ ++dutycounter;
    END_WAVETICKER
};

class triangleunit : public waveunit, public lengthunit
{
    enum {triload, tricount} trimode;
    fBYTE linearcounter, triloadvalue;
    bool oldbit;
    bool jammed;
public:
    fBYTE trianglecounter;
public:
    triangleunit() : lengthunit(7), trimode(triload), linearcounter(0),
                     triloadvalue(0), oldbit(false), jammed(false), trianglecounter(0)
    {
    }
    inline void write(fBYTE ind, fBYTE value)
    {
        //fprintf(stderr, "triangleunit::write(%u, 0x%02X);\n", ind, value);
        waveunit_write(ind, value);
        lengthunit_write(ind, value);
        
        if(ind == 3)
        {
            trimode = (oldbit || !linearcounter) ? triload : tricount;            
        }
        else if(ind == 0)
        {
            triloadvalue = value & 0x7F;
            bool newbit = value & 0x80;
            
            if(!oldbit)
            {
                trimode = tricount;
                linearcounter = triloadvalue;
            }
            else if(!newbit && !linearcounter)
                trimode = tricount;
            
            newbit = oldbit;
        }
        jammed = !linearcounter || getwavelen() < 8;
    }
    void hztick(BYTE ticktype)
    {
        lengthunit_hztick(ticktype);
        
        if(PAPUTICK_240HZ(ticktype))
        {
            if(trimode==tricount && linearcounter > 0)
            {
                --linearcounter;
                if(!linearcounter)jammed = true;
            }
        }
    }
    inline bool silent() const
    {
        return jammed || lengthunit_silent();
    }
    
    BEGIN_WAVETICKER
        /*if(!silent())*/ ++trianglecounter;
    END_WAVETICKER
};

class randgenunit : public decayunit
{
    WORD wlen;
    bool randmode;  // false = 32767-bit, true = 93-bit
    WORD shiftreg;
    bool shifttick;
    unsigned wavepos;
public:
    fBYTE output;
public:
    randgenunit() : wlen(0), shiftreg(1), shifttick(false), wavepos(0), output(0)
    {
    }
    inline void write(fBYTE ind, fBYTE value)
    {
    //    fprintf(stderr, "randgenunit::write(%u, 0x%02X);\n", ind, value);
        decayunit_write(ind, value);
        
        if(ind == 2)
        {
            static const WORD wlenconvdata[16] =
            {0x002, 0x004, 0x008, 0x010,
             0x020, 0x030, 0x040, 0x050,
             0x065, 0x07F, 0x0BE, 0x0FE,
             0x17D, 0x1FC, 0x3F9, 0x7F2};
            wlen = wlenconvdata[value & 15];
            randmode = value >> 7;
        }
    }
    inline bool silent() const
    {
        return lengthunit_silent();
    }
    inline void hztick(BYTE ticktype)
    {
        decayunit_hztick(ticktype);
    }
    inline void waveunit_tick()
    {
    //  This requires CPU tick
        if(++wavepos >= wlen)
        {
            wavepos = 0;
               
            if(shifttick)
            {
                //if(!silent())
                //{
                    output = ((shiftreg >> 14) ^ (shiftreg >> (randmode ? 8 : 13)) ^ 1) & 1;
                    shiftreg = (shiftreg << 1) | output;
                    shifttick = false;
                //}
            }
            else
                shifttick = true;
        }
    }
};

template<class T>
class papuhzticker : public T
{
private:
    unsigned countermax, hzcounter;
    fBYTE hzpos;
public:
    papuhzticker() : hzpos(0)
    {
        countermax = Clock->ClockHz / 240;
        hzcounter = countermax;
    }
    void tick()
    {
        waveunit_tick();
        if(!--hzcounter)
        {
            hzcounter = countermax;
            hztick(++hzpos);
        }
    }
};

typedef papuhzticker<squaregenunit> squaregen;
class sq1gen : public squaregen { public: sq1gen() { set_decr_mode(0); } };
class sq2gen : public squaregen { public: sq2gen() { set_decr_mode(1); } };

class dmcgen
{
	BYTE playbackmode;
    bool enabled;
    fWORD wlen, wcount;       // waveform position counter
    fWORD playaddr, loadaddr; // playing address
    fWORD dmalen, lengthleft; // playing length
public:
	bool irqflag;
    BYTE deltacounter;
public:
    dmcgen() : playbackmode(0), enabled(false), wcount(0xFFF),
               playaddr(0xC000), loadaddr(0xC000), lengthleft(0),
               irqflag(false), deltacounter(0)
    {
    }
    void write(fBYTE ind, fBYTE value)
    {
        fprintf(stderr, "dmc::write(%u, 0x%02X);\n", ind, value);
    	if(!ind)
    	{
            static const WORD wlenconvdata[16] =
            {0xD60, 0xBE0, 0xAA0, 0xA00,
             0x8F0, 0x7F0, 0x710, 0x6B0,
             0x5F0, 0x500, 0x470, 0x400,
             0x350, 0x2A0, 0x240, 0x1B0};
             
            wlen = wlenconvdata[value & 15];
			playbackmode = value >> 6;
			irqflag = false;
			fprintf(stderr, "->wlen %X, mode %d, irq flag reset\n", wlen, playbackmode);
    	}
    	else if(ind==1)
    	{
    		deltacounter = value & 127;
    		fprintf(stderr, "->deltacounter <- %u\n", value & 127);
    	}
    	else if(ind==2)
    	{
    		loadaddr = ((fWORD)value << 6) | 0xC000;
    		playaddr = loadaddr;
    		fprintf(stderr, "->loading address <- %X\n", loadaddr);
    	}
    	else if(ind==3)
    	{
    		dmalen = (fWORD)value << 4;
    		lengthleft = dmalen;
    		fprintf(stderr, "->length <- %u\n", dmalen);
    	}
    }
    inline void enable()
    {
        enabled = true;
        irqflag = false;
    }
    inline void disable()
    {
        enabled = false;
    }
    inline bool silent() const
    {
    	return !enabled || !lengthleft;
    }
    bool tick()
    {
    	if(silent())return false;
    	
	    //  This requires CPU tick
	    if(!wcount)
	    {
			if(wlen)return false;
			wcount = wlen;
		}
    	--wcount;
    	
    	BYTE byt = Op6502(playaddr);
    	if(playaddr == 0xFFFF)playaddr = 0x8000;else ++playaddr;
    	
    	for(unsigned a=0; a<8; ++a)
    	{
    		if(byt&1)
    		{
    			if(deltacounter < 126)deltacounter += 2;
    		}
    		else if(deltacounter > 1)deltacounter -= 2;
	    	byt >>= 1;
	    }
    	
    	if(!--lengthleft)
    	{
    		if(playbackmode==2)
    		{
    			irqflag = true;
    			return true;
    		}
    		if(playbackmode & 1)
    		{
    			lengthleft = dmalen;
    			playaddr = loadaddr;
    		}
   		}
    	return false;
    }
};

/*

Yhteenveto:

 Vasemmampana oleva tarvitsee AINA sisempn (alempana/oikeampana) olevan
 palveluita, ja waveunit tarvitsee aina emonsa wavetick()-funktiota.
 
    |-sq1gen
    |    `- hzticker
    |        `- squaregenunit (generaattoriyksikk)
    |            |- waveunit
    |            `- decayunit
    |                `- lengthunit
    |-sq2gen
    |     `- hzticker
    |        `- squaregenunit (generaattoriyksikk)
    |            |- waveunit
    |            `- decayunit
    |                `- lengthunit
    |-trigen
    |    `- hzticker
    |        `- trianglegenunit (generaattoriyksikk)
    |            |- waveunit
    |            `- lengthunit
    |-rndgen
    |    `- hzticker
    |        `- randgenunit (generaattoriyksikk)
    |            `- decayunit
    |                `- lengthunit
    `-dmcgen
        `- sndchan

waveunit pitisi siis kirjoittaa ilman virtualia.
Miten kaksisuuntainen perint pitisi kirjoittaa?

*/

#endif
